#!/usr/bin/env python3
import struct
from . import base

class ParticleSource:
    def __init__(self, version, pType):
        self.version = version
        self.type = pType
        self.attributes = []

class Particles:
    def __init__(self, version, flags):
        self.version = version
        self.flags = flags
        self.sources = []
    
    def addSource(self, source):
        self.sources.append(source)
    
    def __repr__(self):
        return "<Particles {} flags=0x{:x} sources={}>".format(self.version, self.flags, len(self.sources))
    
    @classmethod
    def loadsVXN(cls, data):
        magic, version, flags = struct.unpack_from("<4sBI", data)
        if magic != b"FvXn":
            raise ValueError("Invalid VXN file!")
        
        if version != 1:
            raise ValueError("Only know version 1")
        
        offset = 9
        
        result = cls(version, flags)
        while offset < len(data):
            magic, pType, version = struct.unpack_from("<2sIB", data, offset)
            if magic != b"m!":
                raise ValueError("Unexpected data in particle stream!")
            
            offset += 7
            source = ParticleSource(version, pType)
            
            sdata = offset
            
            for p in struct.unpack_from("<4Iii5d", data, offset):
                source.attributes.append(p)
            
            offset += 64
            
            if pType == 1:
                for p in struct.unpack_from("<i12d", data, offset):
                    source.attributes.append(p)
                offset += 100
            elif pType == 2:
                for p in struct.unpack_from("<4d", data, offset):
                    source.attributes.append(p)
                offset += 32
            else:
                raise ValueError("Unknown particle type {}".format(pType))
            
            for p in struct.unpack_from("<28dii", data, offset):
                source.attributes.append(p)
            
            offset += 232
            
            result.sources.append(source)
            
        return result
    
    @classmethod
    def loadsMessage(cls, data):
        offset = 0
        if data[:6] == b"VXNASC":
            offset = 6 #Raw, no offsets
        else:
            raise ValueError("Invalid VXN ASCII stream!")
        
        version = base.b220decode(data[offset:offset+1])
        offset += 1
        
        flags = base.b220decode(data[offset:offset+4])
        offset += 4
        
        result = cls(version, flags)
        while offset < len(data):
            if data[offset:offset+2] != b"m!":
                raise ValueError("Unexpected data in particle stream!")
            
            offset += 2
            
            pType = base.b220decode(data[offset:offset+4])
            offset += 4
            
            version = base.b220decode(data[offset:offset+1])
            offset += 1
            
            if pType == 1:
                sourceDataLen = 400
            elif pType == 2:
                sourceDataLen = 328
            else:
                raise ValueError("Unknown source data type!")
            
            sourceData = []
            for i in range(0, sourceDataLen, 1):
                sourceData.append(base.b220decode(data[offset:offset+2]))
                offset += 2
            
            sourceData = bytes(sourceData)
            
            source = ParticleSource(version, pType)
            
            sourceOffset = 0
            
            for p in struct.unpack_from("<4Iii5d", sourceData, sourceOffset):
                source.attributes.append(p)
            
            sourceOffset += 64
            
            if pType == 1:
                for p in struct.unpack_from("<i12d", sourceData, sourceOffset):
                    source.attributes.append(p)
                sourceOffset += 100
            elif pType == 2:
                for p in struct.unpack_from("<4d", sourceData, sourceOffset):
                    source.attributes.append(p)
                sourceOffset += 32
            else:
                raise ValueError("Unknown particle type {}".format(pType))
            
            for p in struct.unpack_from("<28dii", sourceData, sourceOffset):
                source.attributes.append(p)
            
            result.sources.append(source)
        
        return result
    
    def dumpsVXN(self):
        data = b""
        data += struct.pack("<4sBI", b"FvXn", self.version, self.flags)
        
        for source in self.sources:
            data += struct.pack("<2sIB", b"m!", source.type, source.version)
            data += struct.pack("<4Iii5d", *source.attributes[:11])
            offset = 11
            
            if source.type == 1:
                data += struct.pack("<i12d", *source.attributes[offset:offset+13])
                offset += 13
            
            elif source.type == 2:
                data += struct.pack("<4d", *source.attributes[offset:offset+4])
                offset += 4
            
            data += struct.pack("<28dii", *source.attributes[offset:offset+30])
            
        return data
    
    def dumpsTxt(self):
        data = "HORUS VXN TEXT\n"
        data += "version {}\n".format(self.version)
        data += "flags 0b{:0>32b}\n".format(self.flags)
        data += "start sources\n"
        i = 0
        for source in self.sources:
            data += "    start source {}\n".format(i)
            data += "        version {}\n".format(source.version)
            data += "        type {}\n".format(source.type)
            data += "        start attributes\n"
            ii = 0
            for attribute in source.attributes:
                attrName = "attr{}".format(ii)
                
                #Resolve the name if we have it
                if ii in particleAttributeMetadata[source.type]["names"]:
                    attrName = particleAttributeMetadata[source.type]["names"][ii]
                
                if type(attribute) == int:
                    data += "            {} {}\n".format(attrName, attribute)
                elif type(attribute) == float:
                    data += "            {} {}\n".format(attrName, attribute)
                else:
                    print("Invalid attribute value!")
                    return None
                
                ii += 1
            
            data += "        end attributes\n"
            data += "    end source\n"
            i += 1
        data += "end sources\n"
        
        return data.strip().encode()
    
    def dumpsMessage(self):
        data = b"VXNASC" 
        
        data += base.b220encode(self.version, 1)
        data += base.b220encode(self.flags, 4)
        
        for source in self.sources:
            data += b"m!"
            
            data += base.b220encode(source.type, 4)
            data += base.b220encode(source.version, 1)
            
            sourceData = list(struct.pack("<4Iii5d", *source.attributes[:11]))
            offset = 11
            
            if source.type == 1:
                sourceData += list(struct.pack("<i12d", *source.attributes[offset:offset+13]))
                offset += 13
            
            elif source.type == 2:
                sourceData += list(struct.pack("<4d", *source.attributes[offset:offset+4]))
                offset += 4
            
            sourceData += list(struct.pack("<28di", *source.attributes[offset:offset+30]))
            
            for byte in sourceData:
                data += base.b220encode(byte, 2)
        
        return data



#!/usr/bin/env python3

def b95encode(i, size = None):
    out = b""
    while i > 0:
        out = bytes([i % 95 + 32]) + out
        i = i // 95
    
    #Pad if needed
    if size:
        out = out[:size]
        if len(out) < size:
            out = b" "*(size - len(out)) + out
    
    return out

def b95decode(data):
    out = 0
    for c in data:
        if c >= 32 and c < 127:
            out = (out * 95) + (c - 32)
        else:
            raise ValueError("Invalid base95 character!")
    return out

def b220encode(i, size = None):
    out = b""
    while i > 0:
        out = out + bytes([i % 220 + 35])
        i = i // 220
    
    #Pad if needed
    if size:
        out = out[:size]
        if len(out) < size:
            out = out + b"#"*(size - len(out))
    
    return out

def b220decode(data):
    out = 0
    i = 1
    for c in data:
        if c >= 35 and c < 255:
            out = out + ((c - 35) * i)
            i *= 220
        else:
            raise ValueError("Invalid base220 character!")
    return out

def unitTests():
    b95Tests = [
        ((0, None), b''),
        ((94, None), b'~'),
        ((95, None), b'! '),
        ((43, None), b'K'),
        ((583, None), b'&-'),
        ((2299, None), b'83'),
        ((0, 2), b'  '),
        ((94, 2), b' ~'),
        ((95, 2), b'! '),
        ((43, 2), b' K'),
        ((583, 2), b'&-'),
        ((2299, 2), b'83'),
    ]
    for test in b95Tests:
        assert b95encode(*test[0]) == test[1], "Base95 encode for {} (Padding: {}) failed!".format(*test[0])

    for test in b95Tests:
        assert b95decode(test[1]) == test[0][0], "Base95 decode for {} (Padding: {}) failed!".format(*test[0])

    b220Tests = [
        ((0, None), b''),
        ((219, None), b'\xfe'),
        ((220, None), b'#$'),
        ((43, None), b'N'),
        ((583, None), b'\xb2%'),
        ((2299, None), b'\x86-'),
        ((0, 2), b'##'),
        ((219, 2), b'\xfe#'),
        ((220, 2), b'#$'),
        ((43, 2), b'N#'),
        ((583, 2), b'\xb2%'),
        ((2299, 2), b'\x86-'),
    ]
    for test in b220Tests:
        assert b220encode(*test[0]) == test[1], "Base220 encode for {} (Padding: {}) failed!".format(*test[0])

    for test in b220Tests:
        assert b220decode(test[1]) == test[0][0], "Base220 decode for {} (Padding: {}) failed!".format(*test[0])


if __name__ == "__main__":
    unitTests()


#!/usr/bin/env python3
try:
    from PIL import Image, ImageDraw, ImageOps
except ModuleNotFoundError:
    pass

try:
    from libfurc.crypto.fox5 import Fox5Cipher
    haveCrypto = True
except ModuleNotFoundError:
    haveCrypto = False

import io
import lzma
import struct
import random

sUInt64 = struct.Struct("<Q")
def Compress(data):
    compressed = lzma.compress(data, format=lzma.FORMAT_ALONE)
        
    compressed = compressed[:5] + sUInt64.pack(len(data)) + compressed[13:]
    return compressed

def Decompress(data):
    return lzma.decompress(data, format=lzma.FORMAT_ALONE)
    
class Fox5Error(Exception):
    pass

class Fox5Image:
    def __init__(self, width = 0, height = 0, format = 0, data = None):
        self.width = width
        self.height = height
        self.format = format
        self.__data__ = data
        self.compressed = None
    
    def __repr__(self):
        return "<{} Width={}; Height={}; Format={}>".format(self.__class__.__name__, self.width, self.height, self.format)
    
    def __len__(self):
        return len(self.__data__)
    
    @property
    def data(self):
        return self.__data__
    
    @data.setter
    def data(self, value):
        self.__data__ = data
        self.compressed = None
    
    def compress(self):
        if not self.compressed:
            self.compressed = Compress(self.data)
        return self.compressed
    
    @classmethod
    def fromPIL(cls, im):
        data = b""
        for pixel in im.getdata():
            data += bytes((pixel[3], pixel[0], pixel[1], pixel[2]))
        
        return cls(im.width, im.height, 1, data)
    
    def toPIL(self):
        if self.format == 1:
            data = b""
            for pixel in range(0, len(self.data), 4):
                data += bytes((self.data[pixel+1], self.data[pixel+2], self.data[pixel+3], self.data[pixel]))
            return Image.frombytes('RGBA', (self.width, self.height), data)
        elif self.format == 2:
            return Image.frombytes('L', (self.width, self.height), self.data)

sList = struct.Struct(">BI")
sUInt8 = struct.Struct(">B")
sUInt16 = struct.Struct(">H")
sInt16 = struct.Struct(">h")
sUInt32 = struct.Struct(">I")
sInt32 = struct.Struct(">i")
sImageList = struct.Struct(">IHHB")

class Fox5List(object):
    __properties__ = ()
    __childtype__ = None
    properties = {}
    def __init__(self, parent = None):
        self.properties = {}
        for prop in self.__properties__:
            self.properties[prop] = None
        
        self.children = []
        self.childrenLevel = 0
        self.parent = parent
    
    def toDict(self):
        return {
            **self.properties,
            "children": [child.toDict() for child in self.children]
        }
    
    def __getitem__(self, name):
        if type(name) == int:
            return self.children[name]
        elif type(name) == str and name in self.__properties__:
            return self.properties[name]
        else:
            raise Fox5Error("Unknown key {}".format(name))
    
    def __setitem__(self, name, value):
        if type(name) == int:
            self.children[name] = value
        
        elif type(name) == str and name in self.__properties__:
            self.properties[name] = value
        
        else:
            raise Fox5Error("Unknown key {}".format(name))
    
    def __delitem__(self, name):
        if type(name) == int:
            del self.children[name]
        
        elif type(name) == str and name in self.__properties__:
            self.properties.pop(name)
        
        else:
            raise Fox5Error("Unknown key {}".format(name))
    
    def append(self, value):
        self.children.append(value)
    
    def decodeProperty(self, char, handle):
        return False
    
    def encodeProperty(self, prop):
        return False
    
    @classmethod
    def fromStream(cls, handle, *args, **kwargs):
        self = cls(*args, **kwargs)
        char = True
        while True:
            char = handle.read(1)
            if not char:
                break
            
            if char == b"<":
                break
            
            elif char == b"L":
                if self.__childtype__ == None:
                    raise Fox5Error("Unexpected list!")
                
                level, count = sList.unpack(handle.read(sList.size))
                
                if level != self.__childtype__.__level__:
                    raise Fox5Error("Invalid level encountered!")
                
                for i in range(count):
                    self.children.append(self.__childtype__.fromStream(handle, self))
                
                if i+1 != count:
                    raise Fox5Error("Unexpected end of list!")
                
            elif not self.decodeProperty(char, handle):
                    raise Fox5Error("Unhandled attribute {}".format(char))
        return self
    
    def __bytes__(self):
        data = b""
        for prop in self.__properties__:
            if prop in self.properties and self.properties[prop] != None:
                encoded = self.encodeProperty(prop)
                if type(encoded) != bytes:
                    raise Fox5Error("Unhandled attribute {}".format(prop))
                data += encoded
        
        if self.__childtype__ != None:
            data += b"L" + sList.pack(self.__childtype__.__level__, len(self.children))
            for child in self.children:
                data += bytes(child)
                data += b"<"
        return data
        
    
class Fox5Sprite(Fox5List):
    __level__ = 4
    __properties__ = ("Purpose", "Image", "Offset")
    __childtype__ = None
    
    def encodeProperty(self, prop):
        data = b""
        if prop == "Purpose":
            data += b"C" + sUInt16.pack(self["Purpose"])
            return data
        
        if prop == "Image":
            data += b"c" + sUInt16.pack(self["Image"])
            return data
        
        if prop =="Offset":
            data += b"O" + sUInt16.pack(self["Offset"][0]) + sUInt16.pack(self["Offset"][1])
            return data
    
    def decodeProperty(self, char, handle):
        if char == b"C":
            self["Purpose"], = sUInt16.unpack(handle.read(sUInt16.size))
            return True
        
        if char == b"c":
            self["Image"], = sUInt16.unpack(handle.read(sUInt16.size))
            return True
        
        if char == b"O":
            self["Offset"] = [
                sUInt16.unpack(handle.read(sUInt16.size))[0],
                sUInt16.unpack(handle.read(sUInt16.size))[0]
            ]
            return True

class Fox5Frame(Fox5List):
    __level__ = 3
    __properties__ = ("FrameOffset", "FurreOffset")
    __childtype__ = Fox5Sprite
    
    def encodeProperty(self, prop):
        data = b""
        if prop == "FrameOffset":
            data += b"o" + sInt16.pack(self["FrameOffset"][0]) + sInt16.pack(self["FrameOffset"][1])
            return data
        
        if prop == "FurreOffset":
            data += b"f" + sInt16.pack(self["FurreOffset"][0]) + sInt16.pack(self["FurreOffset"][1])
            return data
    
    def decodeProperty(self, char, handle):
        if char == b"o":
            self["FrameOffset"] = [
                sInt16.unpack(handle.read(sInt16.size))[0],
                sInt16.unpack(handle.read(sInt16.size))[0]
            ]
            return True
        
        if char == b"f":
            self["FurreOffset"] = [
                sInt16.unpack(handle.read(sInt16.size))[0],
                sInt16.unpack(handle.read(sInt16.size))[0]
            ]
            return True
            

class Fox5Shape(Fox5List):
    __level__ = 2
    __properties__ = ("Purpose", "Direction", "State", "Ratio",
                      "KitterSpeak")
    __childtype__ = Fox5Frame
    
    def encodeProperty(self, prop):
        data = b""
        if prop == "Purpose":
            data += b"p" + sUInt8.pack(self["Purpose"])
            return data
        
        if prop == "State":
            data += b"s" + sUInt8.pack(self["State"])
            return data
        
        if prop == "Direction":
            data += b"D" + sUInt8.pack(self["Direction"])
            return data
        
        if prop == "Ratio":
            data += b"R" + sUInt8.pack(self["Ratio"][0]) + sUInt8.pack(self["Ratio"][1])
            return data
        
        if prop == "KitterSpeak":
            data += b"K" + sUInt16.pack(len(self["KitterSpeak"]))
            for kline in self["KitterSpeak"]:
                kline = (list(kline) + [0,0,0])[0:3]
                data += sUInt16.pack(kline[0]) + sInt16.pack(kline[1]) + sInt16.pack(kline[2])
            return data
    
    def decodeProperty(self, char, handle):
        if char == b"p":
            self["Purpose"], = sUInt8.unpack(handle.read(sUInt8.size))
            return True
        
        if char == b"s":
            self["State"], = sUInt8.unpack(handle.read(sUInt8.size))
            return True
        
        if char == b"D":
            self["Direction"], = sUInt8.unpack(handle.read(sUInt8.size))
            return True
        
        if char == b"R":
            self["Ratio"] = [
                sUInt8.unpack(handle.read(sUInt8.size))[0],
                sUInt8.unpack(handle.read(sUInt8.size))[0]
            ]
            return True
        
        if char == b"K":
            self["KitterSpeak"] = []
            count, = sUInt16.unpack(handle.read(sUInt16.size))
            for i in range(count):
                self["KitterSpeak"].append([
                    sUInt16.unpack(handle.read(sUInt16.size))[0],
                    sInt16.unpack(handle.read(sInt16.size))[0],
                    sInt16.unpack(handle.read(sInt16.size))[0],
                ])
            return True

class Fox5Object(Fox5List):
    __level__ = 1
    __properties__ = ("ID", "Name", "Description", "Authors", "Revisions",
                      "Keywords", "License", "Portal", "EditType", "Flags",
                      "MoreFlags", "FxFilter")
    __childtype__ = Fox5Shape
    
    def encodeProperty(self, prop):
        data = b""
        if prop == "Revisions":
            data += b"r" + sUInt16.pack(self["Revisions"])
            return data
        
        if prop == "Authors":
            data += b"a" + sUInt16.pack(len(self["Authors"]))
            for i in range(len(self["Authors"])):
                name = self["Authors"][i].encode()
                data += sUInt16.pack(len(name)) + name
            return data
        
        if prop == "License":
            data += b"l" + sUInt8.pack(self["License"])
            return data
        
        if prop == "Keywords":
            data += b"k" + sUInt16.pack(len(self["Keywords"]))
            for i in range(len(self["Keywords"])):
                name = self["Keywords"][i].encode()
                data += sUInt16.pack(len(name)) + name
            return data
        
        if prop == "Name":
            name = self["Name"].encode()
            data += b"n" + sUInt16.pack(len(name)) + name   
            return data
        
        if prop == "Description":
            name = self["Description"].encode()
            data += b"d" + sUInt16.pack(len(name)) + name
            return data
        
        if prop == "Flags":
            data += b"!" + sUInt8.pack(self["Flags"])
            return data
        
        if prop == "Portal":
            name = self["Portal"].encode("iso-8859-1")
            data += b"P" + sUInt16.pack(len(name)) + name
            return data

        if prop == "MoreFlags":
            data += b"?" + sUInt32.pack(self["MoreFlags"])
            return data
        
        if prop == "ID":
            data += b"i" + sInt32.pack(self["ID"])
            return data
        
        if prop == "EditType":
            data += b"t" + sUInt8.pack(self["EditType"])
            return data
        
        if prop == "FxFilter":
            data += b"F" + sUInt8.pack(self["FxFilter"]["Layer"]) + sUInt8.pack(self["FxFilter"]["Mode"])
            return data
    
    def decodeProperty(self, char, handle):
        if char == b"r":
            self["Revisions"], = sUInt16.unpack(handle.read(sUInt16.size))
            return True
        
        if char == b"a":
            self["Authors"] = []
            count, = sUInt16.unpack(handle.read(sUInt16.size))
            for i in range(count):
                l, = sUInt16.unpack(handle.read(sUInt16.size))
                self["Authors"].append(handle.read(l).decode())
            return True
        
        if char == b"l":
            self["License"], = sUInt8.unpack(handle.read(sUInt8.size))
            return True
        
        if char == b"k":
            self["Keywords"] = []
            count, = sUInt16.unpack(handle.read(sUInt16.size))
            for i in range(count):
                l, = sUInt16.unpack(handle.read(sUInt16.size))
                self["Keywords"].append(handle.read(l).decode())
            return True
        
        if char == b"n":
            l, = sUInt16.unpack(handle.read(sUInt16.size))
            self["Name"] = handle.read(l).decode()
            return True
        
        if char == b"d":
            l, = sUInt16.unpack(handle.read(sUInt16.size))
            self["Description"] = handle.read(l).decode()
            return True
        
        if char == b"!":
            self["Flags"], = sUInt8.unpack(handle.read(sUInt8.size))
            return True
        
        if char == b"P":
            l, = sUInt16.unpack(handle.read(sUInt16.size))
            self["Portal"] = handle.read(l).decode("iso-8859-1")
            return True

        if char == b"?":
            self["MoreFlags"], = sUInt32.unpack(handle.read(sUInt32.size))
            return True
        
        if char == b"i":
            self["ID"], = sInt32.unpack(handle.read(sInt32.size))
            return True
        
        if char == b"t":
            self["EditType"], = sUInt8.unpack(handle.read(sUInt8.size))
            return True
        
        if char == b"F":
            self["FxFilter"] = {
                "Layer": sUInt8.unpack(handle.read(sUInt8.size))[0],
                "Mode": sUInt8.unpack(handle.read(sUInt8.size))[0]
            }
            return True

class Fox5Body(Fox5List):
    __level__ = 0
    __properties__ = ("ImageList", "Generator")
    __childtype__ = Fox5Object
    
    def encodeProperty(self, prop):
        data = b""

        if prop == "Generator":
            data += b"g" + sUInt8.pack(self["Generator"])
            return data

        if prop == "ImageList":
            data += b"S" + sUInt32.pack(len(self["ImageList"]))
            for image in self["ImageList"]:
                data += sImageList.pack(len(image.compress()), image.width, image.height, image.format)
                #Encryption is handled in Fox5File
            return data
        
    def decodeProperty(self, char, handle):
        if char == b"g":
            self["Generator"], = sUInt8.unpack(handle.read(sUInt8.size))
            return True

        if char == b"S":
            self["ImageList"] = []
            
            foxhandle = self.parent.parent
            
            count, = sUInt32.unpack(handle.read(sUInt32.size))
            for count in range(count):
                compressedSize, width, height, fmt = sImageList.unpack(handle.read(sImageList.size))
                data = foxhandle.read(compressedSize)
                
                #Decrypt if needed
                if self.parent.encryption == 1:
                    uncompressedSize = width * height * (4 if fmt == 1 else 1)
                    data = Fox5Cipher(data, uncompressedSize, self.parent.seed)
                
                im = Fox5Image(width, height, fmt, Decompress(data))
                self["ImageList"].append(im)
            return True

sFox5Footer = struct.Struct(">BBxxII8s")
class Fox5File(Fox5List):
    __level__ = -1
    __childtype__ = Fox5Body
    
    def __init__(self, *args, compression = 2, encryption = 0, seed = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.compression = compression
        self.encryption = encryption
        self.seed = seed
    
    @property
    def body(self):
        return self.children[0]
    
    @classmethod
    def fromStream(cls, handle):
        handle.seek(-8, 2)
        if handle.read(4) != b"FOX5":
            raise Fox5Error("Not a Fox5 File!")
        
        if handle.read(4) != b".1.1":
            raise Fox5Error("Fox5 file, but no idea what version it is!")
        
        handle.seek(-20, 2)
        compression, encryption, dataCompressedSize, \
            dataUncompressedSize, magic = sFox5Footer.unpack(handle.read(sFox5Footer.size))
        
        if encryption == 1:
            if haveCrypto == False:
                #Only need to check for this once, it won't magically appear.
                raise Fox5Error("Can't decrypt without libfurc.crypto!")
            handle.seek(-36, 2)
            seed = handle.read(16)
        else:
            seed = None
        
        handle.seek(0)
        data = handle.read(dataCompressedSize)
        
        if encryption == 1:
            data = Fox5Cipher(data, dataUncompressedSize, seed)
        
        data = Decompress(data)
        
        if len(data) != dataUncompressedSize:
            raise Fox5Error("Uncompressed size does not equal decompressed size!")
        
        data = io.BytesIO(data)
        if data.read(4) != b"\0\0\0\0":
            raise Fox5Error("Data header is incorrect!")
        
        #This is actually calling the Fox5List
        self = super(Fox5File, cls).fromStream(data, handle, compression = compression, encryption = encryption, seed = seed)
        return self
    
    def __bytes__(self):
        if self.encryption == 1:
            if haveCrypto == False:
                #Only need to check for this once, it won't magically appear.
                raise Fox5Error("Can't encrypt without libfurc.crypto!")
            if self.seed == None:
                seed = random.randbytes(16)
        
        data = b"\0\0\0\0" + super().__bytes__()
        uncompressedSize = len(data)
        data = Compress(data)
        compressedSize = len(data)
        
        if self.encryption == 1:
            data = Fox5Cipher(data, uncompressedSize, self.seed)
        
        for image in self.body["ImageList"]:
            if self.encryption == 1:
                data += Fox5Cipher(image.compress(), len(image), self.seed)
                
            else:
                data += image.compress()
        
        if self.encryption == 1:
            data += self.seed
        
        data += sFox5Footer.pack(self.compression, self.encryption, compressedSize, uncompressedSize, b"FOX5.1.1")
        return data

def load(path):
    with open(path, "rb") as f:
        return Fox5File.fromStream(f)

def loads(data):
    return Fox5File.fromStream(io.BytesIO(data))

def dump(fox, path):
    with open(path, "wb") as f:
        return f.write(bytes(fox))

def dumps(fox):
    return bytes(fox)



#!/usr/bin/env python3

class Colors:
    def __init__(self, version = 116):
        self.version = version
        self.fur = 0
        self.markings = 0
        self.hair = 0
        self.eye = 0
        self.badge = 0
        self.vest = 0
        self.bracer = 0
        self.cape = 0
        self.boot = 0
        self.trousers = 0
        self.wings = 0
        self.accent = 0
        self.gender = None
        self.species = None
        self.avatar = None
    
    def __repr__(self):
        values = ("F: {}; M: {}; E: {}; Ba: {}; V: {}; " \
               + "Br: {}; C: {}; B: {}; T: {}; W: {}; " \
               + "Ac: {}").format(
                   self.fur, self.markings, self.hair, self.hair, self.eye,
                   self.badge, self.vest, self.bracer, self.cape, self.boot,
                   self.trousers, self.wings, self.accent
                )
        if self.gender:
            values += "; G: {}".format(self.gender)
        if self.species:
            values += "; S: {}".format(self.species)
        if self.avatar:
            values += "; Av: {}".format(self.avatar)
        
        return "<Colors({}) {}>".format(self.version, values)
    
    @classmethod
    def fromCode(cls, data):
        self = cls()
        
        return self
    
    @classmethod
    def fromStream(cls, data, readAvatar = True):
        version = data.read(1)[0]
        self = cls(version)
        if version == 116: #t
            if data.remaining < 10:
                raise ValueError("Invalid color code length!")
            
            self.fur = data.read220()
            self.markings = data.read220()
            self.hair = data.read220()
            self.eye = data.read220()
            self.badge = data.read220()
            self.vest = data.read220()
            self.bracer = data.read220()
            self.cape = data.read220()
            self.boot = data.read220()
            self.trousers = data.read220()
            
            if data.remaining >= 1:
                self.gender = data.read220()
            
            if data.remaining >= 1:
                self.species = data.read220()
            
            if data.remaining >= 1:
                self.avatar = data.read220()
            
        elif version == 117: #u
            #TODO: Implement this
            #http://dev.furcadia.com/docs/new_color_code_format.pdf
            if data.remaining < 2:
                raise ValueError("Invalid color code length!")
            
            bitmask = (data.read220()&63) | (data.read220() << 6)
            
            for i in range(0, 10):
                if bitmask >> i & 1:
                    data.read(3)
                else:
                    data.read(1)
            
            if data.remaining >= 1:
                self.gender = data.read220()
            
            if data.remaining >= 1:
                self.species = data.read220()
            
            if data.remaining >= 1:
                self.avatar = data.read220()
        
        elif version == 118: #v
            #TODO: Implement this
            #http://dev.furcadia.com/docs/new_color_code_format.pdf
            if data.remaining < 27:
                raise ValueError("Invalid color code length!")
            
            data.read(27) # 9 * 3
            
            if data.remaining >= 1:
                self.gender = data.read220()
            
            if data.remaining >= 1:
                self.species = data.read220()
            
            if data.remaining >= 1:
                self.avatar = data.read220()
        
        elif version == 119: #w
            if data.remaining < 12:
                raise ValueError("Invalid color code length!")
            
            self.fur = data.read220()
            self.markings = data.read220()
            self.hair = data.read220()
            self.eye = data.read220()
            self.badge = data.read220()
            self.vest = data.read220()
            self.bracer = data.read220()
            self.cape = data.read220()
            self.boot = data.read220()
            self.trousers = data.read220()
            self.wings = data.read220()
            self.accent = data.read220()
            
            if data.remaining >= 1:
                self.gender = data.read220()
            
            if data.remaining >= 1:
                self.species = data.read220()
            
            if data.remaining >= 2 and readAvatar:
                self.avatar = data.read220(2)
            
        else:
            raise ValueError("Invalid color code version {}!".format(version))
        return self



#!/usr/bin/env python3
import struct

class TileAccessor:
    def __init__(self, dream, target, walls = False):
        self.dream = dream
        self.target = target
    
    def __getitem__(self, key):
        if type(key) == int:
            return getattr(self.dream, self.target)[key]
        elif type(key) == tuple and len(key) == 2:
            return getattr(self.dream, self.target)[self.dream.coordinateToIndex(key[0], key[1])]
        else:
            raise ValueError("Unsupported accessor")

    def __setitem__(self, key, value):
        if type(key) == int:
            getattr(self.dream, self.target)[key] = value
        elif type(key) == tuple and len(key) == 2:
            getattr(self.dream, self.target)[self.dream.coordinateToIndex(key[0], key[1])] = value
        else:
            raise ValueError("Unsupported accessor")

class NullAccessor:
    def __init__(self, dream, target, walls = False):
        self.dream = dream
        self.target = target
    
    def __getitem__(self, key):
        return 0

    def __setitem__(self, key, value):
        return 0



sUInt16 = struct.Struct("<H")
sUInt8 = struct.Struct("<B")

DreamTypeCast = {
    'width': int,
    'height': int,
    'revision': int,
    'encoded': bool,
    'patcht': bool,
    'sfxlayermode': str,
    'sfxopacity': int,
    'name': str,
    'patchs': str,
    'noload': bool,
    'allowjs': bool,
    'allowlf': bool,
    'allowfurl': bool,
    'swearfilter': bool,
    'nowho': bool,
    'forcesittable': bool,
    'allowshouts': bool,
    'allowlarge': bool,
    'notab': bool,
    'nonovelty': bool,
    'rating': str,
    'allow32bitart': bool,
    'ismodern': bool
}

class Dream:
    def __init__(self, version = 1.5, **parameters):
        self.properties = {}
        for parameter in parameters:
            if parameter.lower() not in DreamTypeCast:
                raise ValueError("Unknown parameter {}".format(parameter))
            self.properties[parameter.lower()] = DreamTypeCast[parameter.lower()](parameters[parameter])
        
        #Version 100 only supports a static size
        if version == 1.0:
            self.width = 52
            self.height = 100
        
        cells = self.width * self.height
        self._floors = [0] * cells
        self._items = [0] * cells
        self._walls = [0] * cells * 2
        
        self.floors = TileAccessor(self, "_floors")
        self.items = TileAccessor(self, "_items")
        self.walls = TileAccessor(self, "_walls", True)
        
        self._regions = None
        self._effects = None
        
        if version > 1.30:
            self._regions = [0] * cells
            self._effects = [0] * cells
            self.regions = TileAccessor(self, "_regions")
            self.effects = TileAccessor(self, "_effects")
        
        self._lighting = None
        self._ambient = None
        
        if version > 1.50:
            self._lighting = [0] * cells
            self._ambient = [0] * cells
            self._lighting = TileAccessor(self, "_lighting")
            self._ambient = TileAccessor(self, "_ambient")
    
    def coordinateToIndex(self, x, y):
        return (self.height * x) + y
    
    def __getitem__(self, key):
        self.parameters[key]

    def __setitem__(self, key, value):
        if parameter.lower() not in DreamTypeCast:
            raise ValueError("Unknown parameter {}".format(parameter))
        self.parameters[key] = value
    
    @property
    def width(self):
        return self.properties["width"]
    
    @width.setter
    def width(self, value):
        self.properties["width"] = value
    
    @property
    def height(self):
        return self.properties["height"]
    
    @height.setter
    def height(self, value):
        self.properties["height"] = value
    
    @classmethod
    def fromStream(cls, handle):
        header = b""
        while True:
            b = handle.read(1)
            if b == b"\n":
                break
            header += b
        
        if header[0:5] != b"MAP V":
            raise ValueError("Not a map file!")
        
        if header[-9:] != b" Furcadia":
            raise ValueError("Not a map file!")
        
        version = float(header[5:-9])
        
        parameters = {}
        while True:
            param = b""
            while True:
                b = handle.read(1)
                if b == b"\n":
                    break
                param += b
            
            if param == B"BODY":
                break
            
            key, value = param.decode().split("=", 1)
            parameters[key] = value
        
        width, height = int(parameters["width"]), int(parameters["height"])
        
        dream = cls(**parameters, version = version)
        for i in range(0, width*height):
            dream._floors[i], = sUInt16.unpack(handle.read(2))
        
        for i in range(0, width*height):
            dream._items[i], = sUInt16.unpack(handle.read(2))
        
        for i in range(0, width*height*2):
            dream._walls[i], = sUInt8.unpack(handle.read(1))
        
        if version > 1.30:
            for i in range(0, width*height):
                dream._regions[i], = sUInt16.unpack(handle.read(2))
            for i in range(0, width*height):
                dream._effects[i], = sUInt16.unpack(handle.read(2))
        
        if version > 1.50:
            for i in range(0, width*height):
                dream._lighting[i], = sUInt16.unpack(handle.read(2))
            for i in range(0, width*height):
                dream._ambient[i], = sUInt16.unpack(handle.read(2))
        
        return dream

def load(path):
    with open(path, "rb") as f:
        return Dream.fromStream(f)

def loads(data):
    return Dream.fromStream(io.BytesIO(data))

def dump(dream, path):
    with open(path, "wb") as f:
        return f.write(bytes(dream))

def dumps(dream):
    return bytes(dream)




#!/usr/bin/env python3
import struct
import io
import time
import bz2
import binascii
import os
#from . import lzw

#Filename, version, timestamp, filesize, crc32, compressionType
sRCHBlock = struct.Struct("<40sIIIII")

class RCHBlock:
    def __init__(self, parent, filename, version = 1, timestamp = 0, filesize = 0,
                 crc32 = None, compressionType = 2, offset = -1):
        self.parent = parent
        self.name = filename
        self.version = version
        self.timestamp = timestamp
        self.filesize = filesize
        self.crc32 = crc32
        self.compressionType = compressionType
        self.offset = offset
    
    def read(self, uncompressed = False):
        #Seek to location and past header
        self.parent.handle.seek(self.offset + sRCHBlock.size + 2)
        data = self.parent.handle.read(self.filesize)
        if uncompressed:
            return data
        
        if self.compressionType == 0:
            data = bytearray(data)
            for i in range(len(data)):
                data[i] = data[i] ^ 255
            raise NotImplementedError("LZW is not implemented at the moment!")
            decomp = lzw.LzwDecompressor()
            data = decomp.decompress(data)
        elif self.compressionType == 1:
            data = bytearray(data)
            for i in range(len(data)):
                data[i] = data[i] ^ 255
            data = bytes(data)
        elif self.compressionType == 2:
            data = bz2.decompress(data)
        else:
            raise ValueError("Invalid compression type {}".format(self.compressionType))
        return data
    
    def test(self):
        #For some reason, bzip checksum is the uncompressed data
        #all others is the checksum of the compressed data
        if self.compressionType == 2:
            return self.crc32 == binascii.crc32(self.read())
        
        elif self.compressionType == 0 or self.compressionType == 1:
            return self.crc32 == binascii.crc32(self.read(uncompressed=True))
    
#Version, creation time, 4 unsigned ints of reserved
sRCHHeader = struct.Struct("<II16x")

class RCHFile:
    def __init__(self, handle, compressionlevel = 9):
        self.handle = handle
        self.compressionlevel = compressionlevel
        self.files = []
    
    def load(self):
        self.handle.seek(0)
        """Load a metadata into memory"""
        magic = self.handle.read(4)
        
        if len(magic) == 0: #Create data
            return
        elif magic == b"FR01": #Load existing data
            version, timestamp = sRCHHeader.unpack(self.handle.read(sRCHHeader.size))
            #Disabled this check because it'll always be the same thing
            #but the installer has different values.
            #if version != 1:
            #    raise ValueError("Don't know how to read RCH version {}".format(version))
        else:
            raise ValueError("Not a valid RCH archive!")
        
        self.files = []
        
        while True:
            offset = self.handle.tell()
            magic = self.handle.read(2)
            if len(magic) == 0:
                #End of file
                break
            
            if magic != b"FZ":
                raise ValueError("Corrupted RCH file!")
            
            else:
                filename, version, timestamp, filesize, crc32, \
                    compressionType = sRCHBlock.unpack(self.handle.read(sRCHBlock.size))
                
                if version != 1:
                    raise ValueError("Don't know how to read RCH block version {}".format(version))
                
                filename = filename.rstrip(b"\0").decode()
                
                self.files.append(RCHBlock(self, filename, version, timestamp, \
                    filesize, crc32, compressionType, offset))
                self.handle.seek(filesize, 1) #Seek past the data
    
    def getFile(self, name):
        for file in self.files:
            if file.name == name:
                return file
        return None
    
    def add(self, name, data):
        self.handle.seek(0, 2)
        self.handle.write(b"FZ")
        test = bz2.compress(data)
        crc32 = 0
        compression = 1
        if len(data) > len(test):
            #Yes this is correct, before we compress
            crc32 = binascii.crc32(data)
            data = test
            compression = 2
            del test
        else:
            del test
            data = bytearray(data)
            for i in range(len(data)):
                data[i] = data[i] ^ 255
            data = bytes(data)
            #Yes this is correct, after we "compress"
            crc32 = binascii.crc32(data)
        
        self.handle.write(sRCHBlock.pack(name.encode(), 1, 0, len(data), crc32, compression))
        self.handle.write(data)
    
    def write(self):
        self.handle.seek(0)
        self.handle.truncate()
        self.handle.write(b"FR01")
        #Version 1, no timestamp
        self.handle.write(sRCHHeader.pack(1, 0))
        
        chunks = {}
        for file in self.files:
            if file.filesize > 0:
                chunks[file.name] = file.read()
        
        for chunk in chunks:
            self.handle.write(b"FZ")
            data = chunks[chunk]
            test = bz2.compress(data)
            crc32 = 0
            compression = 1
            if len(data) > len(test):
                #Yes this is correct, before we compress
                crc32 = binascii.crc32(data)
                data = test
                compression = 2
                del test
            else:
                del test
                data = bytearray(data)
                for i in range(len(data)):
                    data[i] = data[i] ^ 255
                data = bytes(data)
                #Yes this is correct, after we "compress"
                crc32 = binascii.crc32(data)
            
            self.handle.write(sRCHBlock.pack(chunk.encode(), 1, 0, len(data), crc32, compression))
            self.handle.write(data)
        
        self.handle.truncate()
    
    @classmethod
    def open(cls, filename, mode = "rb"):
        handle = None
        
        if type(filename) == str:
            handle = open(filename, mode)
        elif isinstance(filename, io.IOBase):
            handle = filename
        
        self = cls(handle)
        self.load()
        return self

def main():
    import argparse
    parser = argparse.ArgumentParser(description='Handle RCH files')
    parser.add_argument('command', choices=("a", "d", "e", "l", "t"),
                        help='Command to use. (a = add; d = delete; e = extract')
    parser.add_argument('archive')
    parser.add_argument('filenames', nargs="*")

    args = parser.parse_args()
    
    mode = None
    if args.command in ["a", "d"]:
        mode = "a+b"
    elif args.command in ["e", "l", "t"]:
        mode = "rb"
    
    handle = open(args.archive, mode)
    
    archive = RCHFile.open(handle)
    
    if args.command == "l":
        for file in archive.files:
            print(file.name)
    
    elif args.command == "t":
        for file in archive.files:
            print("[{}] {}".format("PASS" if file.test() else "FAIL", file.name))
    
    elif args.command == "e":
        if not args.filenames:
            args.filenames = [i.name for i in archive.files]
        
        for file in args.filenames:
            data = archive.getFile(file)
            if not data:
                print("Couldn't find file \"{}\" in archive!".format(file))
            with open(file, "wb") as f:
                f.write(data.read())
        
    elif args.command == "d":
        for file in args.filenames:
            data = archive.getFile(file)
            if not data:
                print("Couldn't find file \"{}\" in archive!".format(file))
            data.filesize = -1
        archive.write()
    
    elif args.command == "a":
        for file in args.filenames:
            with open(file, "rb") as f:
                archive.add(os.path.basename(file), f.read())
    

if __name__ == "__main__":
    main()



#!/usr/bin/env python3 
import asyncio
from . import base
from .furcbuffer import FurcBuffer
from .account import Character
from .colors import Colors
from .particles import Particles
import datetime
import time
import math
import logging

LIVE_SERVER = ("lightbringer.furcadia.com", 6500)

#Incoming message definitions
class DefaultPacketHandler:
    async def message_unhandled(self, opcode, data):
        logging.error("Unhandled {}".format(opcode), data)

class PacketHooks(DefaultPacketHandler):
    def __init__(self, useLookups = False):
        self.listeners = {
            "*": []
        }
        self.lookup = None
        if useLookups:
            self.generateLookups()
    
    def generateLookups(self):
        self.lookup = {}
        for entry in dir(self):
            if entry.startswith("message_"):
                try:
                    n = entry.split("_")[1:]
                    if len(n) == 1:
                        n = int(n[0])
                    else:
                        n = (int(n[0]), int(n[1]))
                    self.lookup[n] = getattr(self, entry)
                except ValueError:
                    pass
    
    #Messages
    #" " - Remove avatar by ID
    async def message_0(self, opcode, data):
        msg = FurcBuffer(data)
        fuid = msg.read220(4)
        await self.fire("RemoveAvatarID", fuid)
    
    #"!" - Sound
    async def message_1(self, opcode, data):
        msg = FurcBuffer(data)
        soundID = msg.read95(2)
        await self.fire("Sound", soundID)
    
    #"%" - Standing on object
    async def message_5(self, opcode, data):
        msg = FurcBuffer(data)
        itemID = msg.read95(2)
        await self.fire("ButlerFeet", itemID)
    
    #"&" - Login success
    async def message_6(self, opcode, data):
        await self.fire("Login", True)
    
    #"(" - Text message
    async def message_8(self, opcode, data):
        await self.fire("Message", data)
    
    #")" - Remove avatar
    async def message_9(self, opcode, data):
        msg = FurcBuffer(data)
        fuid = msg.read220(4)
        await self.fire("RemoveAvatar", fuid)
    
    #"/" - Animate avatar
    async def message_15(self, opcode, data):
        msg = FurcBuffer(data)
        fuid = msg.read220(4)
        x = msg.read220(2)
        y = msg.read220(2)
        direction = msg.read220(1)
        shape = msg.read220(1)
        await self.fire("AnimateAvatar", fuid, (x,y), direction, shape)
    
    #"0" - Sync dream variables
    async def message_16(self, opcode, data):
        msg = FurcBuffer(data)
        variables = {}
        while msg.remaining >= 2:
            offset = msg.read95(2)
            while msg.remaining >= 6:
                val = msg.read95(3)
                if val != 0x4000:
                    variables[offset] = val
                    offset += 1
                else:
                    repeats = msg.read95(3) + 1
                    val = msg.read95(3)
                    for i in range(repeats):
                        variables[offset] = val
                        offset += 1
        await self.fire("UpdateVariables", variables)
    
    #"1" - Set floors
    async def message_17(self, opcode, data):
        msg = FurcBuffer(data)
        result = []
        while msg.remaining >= 6:
            x = msg.read220(2)
            y = msg.read220(2)
            floorID = msg.read220(2)
            
            #Decode repeats
            repeats = 48 * math.floor(x / 1000) + math.floor(y / 1000)
            x = x % 1000
            y = y % 1000
            
            result.append({
                "pos": (x, y),
                "id": floorID,
                "repeats": repeats
            })
        
        await self.fire("SetFloor", result)
    
    #"2" - Set walls
    async def message_18(self, opcode, data):
        msg = FurcBuffer(data)
        result = []
        while msg.remaining >= 6:
            x = msg.read220(2)
            y = msg.read220(2)
            wallID = msg.read220(2)
            
            #Decode repeats
            repeats = 48 * math.floor(x / 1000) + math.floor(y / 1000)
            x = x % 1000
            y = y % 1000
            
            result.append({
                "pos": (x, y),
                "id": wallID,
                "repeats": repeats
            })
        await self.fire("SetWall", result)
    
    #"3" - Set all DS variables
    async def message_19(self, opcode, data):
        msg = FurcBuffer(data)
        vals = []
        while msg.remaining >= 3:
            var = msg.read95(3)
            vals.append(var)
        await self.fire("SetVariables", vals)
    
    #"4" - Set region
    async def message_20(self, opcode, data):
        msg = FurcBuffer(data)
        result = []
        while msg.remaining >= 6:
            x = msg.read220(2)
            y = msg.read220(2)
            regionID = msg.read220(2)
            result.append({
                "pos": (x, y),
                "id": regionID
            })
        await self.fire("SetRegion", result)
    
    #"5" - Set effect
    async def message_21(self, opcode, data):
        msg = FurcBuffer(data)
        result = []
        while msg.remaining >= 6:
            x = msg.read220(2)
            y = msg.read220(2)
            effectID = msg.read220(2)
            
            #Decode repeats
            repeats = 48 * math.floor(x / 1000) + math.floor(y / 1000)
            x = x % 1000
            y = y % 1000
            
            result.append({
                "pos": (x, y),
                "id": effectID,
                "repeats": repeats
            })
        await self.fire("SetEffect", result)
    
    #"6" - DS event triggered by player
    async def message_22(self, opcode, data):
        msg = FurcBuffer(data)
        from_x = msg.read95(2)
        from_y = msg.read95(2)
        to_x = msg.read95(2)
        to_y = msg.read95(2)
        while msg.remaining >= 6:
            line = msg.read95(2)
            if line > 8000:
                line = line - 8000 + (msg.read95(2) * 1000)
            
            extra_x = msg.read95(2)
            extra_y = msg.read95(2)
            
            result = {
                "from": (from_x, from_y),
                "to": (to_x, to_y),
                "extra": (extra_x, extra_y),
                "line": line
            }
            await self.fire("DSEvent", True, result)
    
    #"7" - DS event triggered by other
    async def message_23(self, opcode, data):
        msg = FurcBuffer(data)
        from_x = msg.read95(2)
        from_y = msg.read95(2)
        to_x = msg.read95(2)
        to_y = msg.read95(2)
        while msg.remaining >= 8:
            line = msg.read95(2)
            if line > 8000:
                line = line - 8000 + (msg.read95(2) * 1000)
            
            extra_x = msg.read95(2)
            extra_y = msg.read95(2)
        
            result = {
                "from": (from_x, from_y),
                "to": (to_x, to_y),
                "extra": (extra_x, extra_y),
                "line": line
            }
            await self.fire("DSEvent", False, result)
    
    #"8" - DS event addon
    async def message_24(self, opcode, data):
        msg = FurcBuffer(data)
        moveFlag = msg.read95(1)
        randSeed = msg.read95(5)
        try:
            saidNum = msg.read95(3)
        except ValueError:
            #TEMPORARY: Fix for server sending non-base95 numbers
            saidNum = 0
        facingDir = msg.read95(1)
        entryCode = msg.read95(3)
        objPaws =  msg.read95(3)
        furreCount = msg.read95(2)
        userID = msg.read95(6)
        DSBtn = msg.read95(2)
        dreamCookies = msg.read95(3)
        triggererCookies = msg.read95(2)
        portalOpen = (msg.read95(2), msg.read95(2))
        second = msg.read95(1)
        minute = msg.read95(1)
        hour = msg.read95(1)
        day = msg.read95(1)
        month = msg.read95(1)
        year = msg.read95(2)
        portalClose = (msg.read95(2), msg.read95(2))
        await self.fire("DSEventAddon", {
            "moveFlag": moveFlag,
            "randSeed": randSeed,
            "saidNum": saidNum,
            "facingDir": facingDir,
            "entryCode": entryCode,
            "objPaws": objPaws,
            "furreCount": furreCount,
            "userID": userID,
            "DSBtn": DSBtn,
            "dreamCookies": dreamCookies,
            "triggererCookies": triggererCookies,
            "portalOpen": portalOpen,
            "second": second,
            "minute": minute,
            "hour": hour,
            "day": day,
            "month": month,
            "year": year,
            "portalClose": portalClose
        })
    
    #"9" - Region flags
    async def message_25(self, opcode, data):
        msg = FurcBuffer(data)
        regions = {}
        while msg.remaining >= 8:
            index = msg.read220(3)
            flag = msg.read220(3)
            count = msg.read220(2)
            for i in range(count):
                regions[0xFFFF & i] = flag
                
        await self.fire("RegionFlags", regions)
    
    #";" - Load map
    async def message_27(self, opcode, data):
        await self.fire("LoadMap", data.decode())
    
    #"<" - Spawn avatar
    async def message_28(self, opcode, data):
        msg = FurcBuffer(data)
        fuid = msg.read220(4)
        x = msg.read220(2)
        y = msg.read220(2)
        direction = msg.read220(1)
        shape = msg.read220(1)
        nameLength = msg.read220(1)
        name = msg.read(nameLength)
        colors = Colors.fromStream(msg, False)
        flags = msg.read220(1)
        afkTime = msg.read220(4)
        scale = msg.read220(1)
        await self.fire("SpawnAvatar", fuid, (x,y), direction, shape, name, colors, flags, afkTime, scale)
    
    #"=" - Resume drawing
    async def message_29(self, opcode, data):
        await self.fire("Suspend", False)
    
    #">" - Spawn object
    async def message_30(self, opcode, data):
        msg = FurcBuffer(data)
        result = []
        while msg.remaining >= 6:
            x = msg.read220(2)
            y = msg.read220(2)
            objectID = msg.read220(2)
            
            #Decode repeats
            repeats = 48 * math.floor(x / 1000) + math.floor(y / 1000)
            x = x % 1000
            y = y % 1000
            
            result.append({
                "pos": (x, y),
                "id": objectID,
                "repeats": repeats
            })
        
        await self.fire("SetObject", result)
    
    #"@" - Move camera
    async def message_32(self, opcode, data):
        msg = FurcBuffer(data)
        x = msg.read95(2)
        y = msg.read95(2)
        result = {
            "to": (x, y)
        }
        if msg.remaining >= 4:
            x = msg.read95(2)
            y = msg.read95(2)
            result["from"] = (x, y)
        await self.fire("MoveCamera", result)
    
    #"A" - Move avatars
    async def message_33(self, opcode, data):
        msg = FurcBuffer(data)
        fuid = msg.read220(4)
        x = msg.read220(2)
        y = msg.read220(2)
        direction = msg.read220(1)
        shape = msg.read220(1)
        await self.fire("MoveAvatar", fuid, (x,y), direction, shape)
    
    #"B" - Set avatar colors
    async def message_34(self, opcode, data):
        msg = FurcBuffer(data)
        furre = msg.read220(4)
        direction = msg.read220(1)
        unk1 = msg.read220(1)
        colors = Colors.fromStream(msg)
        await self.fire("SetAvatarColors", furre, direction, unk1, colors)
    
    #"C" - Hide avatar at position
    async def message_35(self, opcode, data):
        msg = FurcBuffer(data)
        furre = msg.read220(4)
        pos = (msg.read220(2), msg.read220(2))
        await self.fire("HideAvatar", furre, pos)
    
    #"D" - Announce furre presence
    async def message_36(self, opcode, data):
        msg = FurcBuffer(data)
        fuid = msg.read220(4)
        if fuid == 0: #Server side bug fix
            return
        x = msg.read220(2)
        y = msg.read220(2)
        direction = msg.read220(1)
        shape = msg.read220(1)
        unk1 = msg.read220(4)
        unk2 = msg.read220(4)
        unk3 = msg.read220(2)
        unk4 = msg.read220(2)
        await self.fire("FurreArrive", fuid, (x, y), direction, shape, unk1, unk2, unk3, unk4)
    
    #"E" - Spawn SFX
    async def message_37(self, opcode, data):
        msg = FurcBuffer(data)
        result = []
        while msg.remaining >= 6:
            x = msg.read220(2)
            y = msg.read220(2)
            fxID = msg.read220(2)
            
            #Decode repeats
            repeats = 48 * math.floor(x / 1000) + math.floor(y / 1000)
            x = x % 1000
            y = y % 1000
            
            result.append({
                "pos": (x, y),
                "id": fxID,
                "repeats": repeats
            })
        
        await self.fire("SetSFX", result)
    
    #"F" - Spawn Ambient
    async def message_38(self, opcode, data):
        msg = FurcBuffer(data)
        result = []
        while msg.remaining >= 6:
            x = msg.read220(2)
            y = msg.read220(2)
            ambientID = msg.read220(2)
            
            #Decode repeats
            repeats = 48 * math.floor(x / 1000) + math.floor(y / 1000)
            x = x % 1000
            y = y % 1000
            
            result.append({
                "pos": (x, y),
                "id": ambientID,
                "repeats": repeats
            })
        
        await self.fire("SetAmbient", result)
    
    #"[" - Disconnected (Reconnect allowed)
    async def message_59(self, opcode, data):
        await self.fire("Disconnect", data.decode())
    
    #"\" - Server authenticate
    async def message_60(self, opcode, data):
        await self.fire("Authenticate", data.decode())
    
    #"]" - Protocol extension
    async def message_61(self, opcode, data):
        if len(data) < 2:
            return
        await getattr(
            self,
            "message_61_" + str(data[0]-32),
            self.message_61_unhandled
        )(data[0]-32, data[1:])
    
    #Default handler
    async def message_61_unhandled(self, opcode, data):
        logging.error("Unhandled 61:{}".format(opcode), data)
    
    #"]!" - Adult warning
    async def message_61_1(self, opcode, data):
        maturity = data.decode()
        parental = False
        if maturity.startswith("+"):
            parental = True
            maturity = maturity[1:]
        
        await self.fire("MaturityWarning", parental, maturity)
    
    #"]#" - Dialog
    async def message_61_3(self, opcode, data):
        msg = FurcBuffer(data)
        dialogID = msg.readUntil().decode()
        
        if dialogID == "xxxx":
            dialogID = -1
        else:
            dialogID = int(dialogID)
        
        dialogType = int(msg.readUntil())
        dialogMessage = msg.read()
        await self.fire("Dialog", dialogID, dialogType, dialogMessage)
    
    #"]$" - Open URL
    async def message_61_4(self, opcode, data):
        msg = FurcBuffer(data)
        url = msg.readUntil()
        await self.fire("OpenURL", False, url)
    
    #"]%" - Online reply
    async def message_61_5(self, opcode, data):
        msg = FurcBuffer(data)
        online = bool(int(msg.read(1)))
        name = msg.read()
        await self.fire("OnlineStatus", online, name)
    
    #"]&" - Set portrait
    async def message_61_6(self, opcode, data):
        msg = FurcBuffer(data)
        pid = int(msg.readUntil().decode())
        name = msg.read() #Portrait name, not furre name!
        await self.fire("Portrait", pid, name)
    
    #"]*" - Show URL
    async def message_61_10(self, opcode, data):
        await self.fire("OpenURL", True, data)
    
    #"]+" - Set last profile ID (DEPRECATED)
    async def message_61_11(self, opcode, data):
        msg = FurcBuffer(data)
        pid = int(msg.readUntil().decode())
        #For use with:
        #http://www.furcadia.com/services/profile/profile.php4?cmd=view&p=<PID>
        await self.fire("LastProfileID", pid)
    
    #"]-" - Prefix text, used by say command (Can be overridden by 61_48)
    async def message_61_13(self, opcode, data):
        await self.fire("PrefixBadge", data)
    
    #"]3" - Show user list
    async def message_61_19(self, opcode, data):
        msg = FurcBuffer(data)
        #0 = Show usercount + list, 1 = Only count
        enabled = not bool(int(msg.read(1)))
        await self.fire("EnableUserList", enabled)
    
    #"]?" - Unknown, something related to pounce
    #FIXME: Figure out wtf this is
    async def message_61_31(self, opcode, data):
        """Seen values:
            char[1] "I" - Clear(?) and jump to func1
            char[1] "A" - jump to func1
            char[1] "S" - jump to func2
            char[1] "o" - jump to func3
            
            func1:
                Base220(4) unk1
                Base220(4) unk2
                char[1] unk3 #flag?
                Base220(4) unk4 #Time?
                Base220(4) unk5 #Time?
                Base220(4) unk6 #Time?
                Base220(1) unk7
                Base220(1) unk8
            
            func2:
                while(remaining > 3)
                    Base220(4) unk1
                    Base220(4) unk2
                    char[1] unk3
                    Base220(4) unk4
                    Base220(4) unk5
            
            func3:
                Base220(4) unk1
                char[1] flag - "+" or "-"
        """
        #await self.fire("SetUserID", int(data.decode()))
    
    
    #"]A" - Download guild tag command (DEPRECATED?)
    async def message_61_33(self, opcode, data):
        msg = FurcBuffer(data)
        checksum = int(msg.readUntil()) #Download it from file server using gt%i
        name = msg.readUntil()
        await self.fire("GuildTagDownload", checksum, name)
    
    #"]B" - Set user ID
    async def message_61_34(self, opcode, data):
        msg = FurcBuffer(data)
        fuid = int(msg.readUntil())
        username = msg.readUntil()
        await self.fire("SetUserID", fuid, username)
    
    #"]C" - Bookmark Dream
    async def message_61_35(self, opcode, data):
        msg = FurcBuffer(data)
        msg = FurcBuffer(data)
        temporary = not bool(int(msg.read(1)))
        fdl = msg.read()
        await self.fire("Bookmark", temporary, fdl)
    
    #"]D" - Begin share edit
    async def message_61_36(self, opcode, data):
        #Activate, Is Owner
        await self.fire("LiveEdit", True, False)
    
    #"]E" - Begin share edit
    async def message_61_37(self, opcode, data):
        #Activate, Is Owner
        await self.fire("LiveEdit", True, True)
    
    #"]F" - Begin share edit
    async def message_61_38(self, opcode, data):
        #Activate, Is Owner
        await self.fire("LiveEdit", False, False)
    
    #"]G" - Disable tab key
    async def message_61_39(self, opcode, data):
        #0 = Enable, 1 = Disable
        msg = FurcBuffer(data)
        disabled = not bool(int(msg.read(1)))
        await self.fire("DisableTab", disabled)
    
    #"]H" - Offset avatar
    async def message_61_40(self, opcode, data):
        msg = FurcBuffer(data)
        fuid = msg.read220(4)
        x = msg.read220(2) - 256
        y = msg.read220(2) - 256
        await self.fire("OffsetAvatar", fuid, (x, y))
    
    #"]I" - Particle Effect
    async def message_61_41(self, opcode, data):
        msg = FurcBuffer(data)
        x = msg.read220(2)
        y = msg.read220(2)
        offset_x = msg.read220(2)
        offset_y = msg.read220(2)
        particles = Particles.loadsMessage(data[8:])
        await self.fire("Particles", (x, y), (offset_x, offset_y), particles)
    
    #"]J" - Web map
    async def message_61_42(self, opcode, data):
        await self.fire("WebMap", data)
    
    #"]K" - Focus DSBtn? Return if visible?
    async def message_61_43(self, opcode, data):
        pass
    
    #"]M" - Dynamic Avatars Info
    async def message_61_45(self, opcode, data):
        avatars = {}
        msg = FurcBuffer(data)
        dataVersion = msg.read220(1)
        if data == 2:
            while msg.remaining >= 8:
                avatarVersion = msg.read220(1)
                
                flags = msg.read220(1)
                
                idP1 = msg.read220(1)
                idP2 = msg.read220(1)
                avatarId = idP1 + 220 * (idP2 >> 1)
                
                updated = msg.read220(4)
                
                avatars[avatarId] = {
                    "version": avatarVersion,
                    "flags": flags,
                    "updated": updated
                }
        
        await self.fire("DynamicAvatars", avatars)
    
    #"]N" - Show pounce channel list
    async def message_61_46(self, opcode, data):
        msg = FurcBuffer(data)
        showPounceChannels = bool(int(msg.read(1)))
        await self.fire("ShowPounceChannels", showPounceChannels)
    
    #"]O" - Gloam
    async def message_61_47(self, opcode, data):
        msg = FurcBuffer(data)
        result = {}
        
        while msg.remaining >= 12:
            fuid = msg.read220(4)
            
            color = 0
            
            for i in range(6):
                #I still don't fully understand this but this seems to work
                color = color << 4
                fragment = msg.read(1)[0]
                if fragment >= 48 and fragment <= 57:
                    color = color | (fragment & 15)
                elif fragment >= 65 and fragment <= 70:
                    color = color | (fragment - 75)
            
            intensity = msg.read220(2)
            
            result[fuid] = {
                "color": color,
                "intensity": intensity
            }
        
        await self.fire("Gloam", result)
    
    #"]P" - Prefix next line (Can be overridden by 61_13)
    async def message_61_48(self, opcode, data):
        await self.fire("PrefixLine", data)
    
    #"]S" - Connection security info
    async def message_61_51(self, opcode, data):
        msg = FurcBuffer(data)
        #0 = insecure, 1 = unk, 2 = unused, 3 = secure
        flag = int(msg.read(1))
        await self.fire("ConnectionSecurity", flag)
    
    #"]W" - Region settings
    #FIXME: Resolve unknown variables
    async def message_61_55(self, opcode, data):
        msg = FurcBuffer(data)
        result = {
            "outdoor": {
                "object": msg.read220(2),
                "wall": msg.read220(2),
                "floor": msg.read220(2),
                "effect": msg.read220(2),
            },
            "indoor": {
                "object": msg.read220(2),
                "wall": msg.read220(2),
                "floor": msg.read220(2),
                "effect": msg.read220(2),
            },
            "unk1": msg.read220(2),
            "walls": {
                "bottom": msg.read220(1),
                "top": msg.read220(1),
            },
            "unk2": msg.read220(2),
            "unk3": msg.read220(2),
            "unk4": msg.read220(2),
            "unk5": msg.read220(2),
            "unk6": []
        }
        
        while msg.remaining >= 4:
            l = msg.read220(1)
            for i in range(l):
                result["unk6"].append((
                    msg.read220(2),
                    msg.read220(1),
                ))
        
        await self.fire("RegionSettings", result)
    
    #"]Z" - Art party(?)
    #TODO: Implement
    async def message_61_58(self, opcode, data):
        """
            Format:
                char[1] opcode # 0 - 7
                
            if opcode == 0: #End
                #No params
            
            if opcode == 2: #Draw
                #Probably base220 byte array
            
            if opcode == 3: #UNKNOWN
                #unknown
            
            if opcode == 4: #Set shape
                base220(4) shape
                
            if opcode == 5:
                #unknown
            
            if opcode == 6: #Request data?
                #no op
            
            if opcode == 7:
                char(*) filename
        """
        pass
    
    #"]]" - Login failure
    async def message_61_61(self, opcode, data):
        await self.fire("Login", False)
    
    #"]_" - Kitterdust
    async def message_61_63(self, opcode, data):
        msg = FurcBuffer(data)
        avatars = {}
        while msg.remaining >= 5:
            avatars[msg.read220(4)] = msg.read220(1)
        await self.fire("KitterDust", avatars)
    
    #"]`" - Load remote config
    async def message_61_64(self, opcode, data):
        msg = FurcBuffer(data)
        subop = msg.read(1)
        if subop == b"e":
            subop = int(msg.read(1))
            if subop == 1:
                afkTime = int(msg.readUntil())
                disconnectTime = int(msg.readUntil())
                await self.fire("RemoteConfig", "Timers", afkTime, disconnectTime)
            
            else:
                #afkconfig <int> <str>
                await self.fire("RemoteConfig", "RequestSend")
        
        elif subop == b"r":
            message = msg.read()
            await self.fire("RemoteConfig", "AutoWhisper", message)
        
        elif subop == b"m":
            message = msg.read()
            await self.fire("RemoteConfig", "AFKWhisper", message)
    
    #"]a" - Upload dream request (DEPRECATED)
    async def message_61_65(self, opcode, data):
        logging.warn("Received deprecated 61:65 (Upload dream request) from server!")
    
    #"]b" - Resize map
    async def message_61_64(self, opcode, data):
        msg = FurcBuffer(data)
        x = msg.read95(2)
        y = msg.read95(2)
        await self.fire("DreamResize", (x, y))
    
    #"]c" - Set marbled
    async def message_61_67(self, opcode, data):
        #This has a "c" prefixed to it
        msg = FurcBuffer(data)
        
        #t is Type
        #type 0 and 1 are unknown
        #type 2 is normal mode
        t = msg.read(1)[0]
        if t > 96 and t < 123:
            t = t - 97
        else:
            t = 0
        
        background = msg.readUntil()
        
        await self.fire("Marbled", t, background)
    
    #"]d" - Something something stickers prototype (DEPRECATED)
    async def message_61_68(self, opcode, data):
        logging.warn("Received deprecated 61:68 (Unknown) from server!")
    
    #"]e" - Something something stickers prototype (DEPRECATED)
    async def message_61_69(self, opcode, data):
        logging.warn("Received deprecated 61:69 (Unknown) from server!")
    
    #"]f" - Look
    async def message_61_70(self, opcode, data):
        msg = FurcBuffer(data)
        colors = Colors.fromStream(msg, False)
        name = msg.read().decode()
        await self.fire("Look", colors, name)
    
    #"]g" - Execute binary (lmao) (DEPRECATED)
    async def message_61_71(self, opcode, data):
        msg = FurcBuffer(data)
        command = msg.read()
        await self.fire("Execute", command, False)
    
    #"]h" - Execute binary and close (lmao)
    async def message_61_72(self, opcode, data):
        msg = FurcBuffer(data)
        command = msg.read()
        await self.fire("Execute", command, True)
    
    #"]i" - Create a file with path name (DEPRECATED)
    async def message_61_73(self, opcode, data):
        """
            Format:
                char(1) bool # 1 = write, anything else is NOP
                char(*) path #?
        """
        logging.warn("Received deprecated 61:73 (Create file) from server!")
    
    #"]j" - Music
    async def message_61_74(self, opcode, data):
        msg = FurcBuffer(data)
        musicID = msg.read95(2)
        await self.fire("Music", musicID)
    
    #"]k" - Set file server
    async def message_61_75(self, opcode, data):
        msg = FurcBuffer(data)
        serverID = msg.read95(1)
        await self.fire("SetFileServer", serverID)
    
    #"]m" - Marco
    async def message_61_77(self, opcode, data):
        msg = FurcBuffer(data)
        msg.readUntil() #Skip space
        polon = msg.read()
        await self.fire("Marco", polon)
    
    #"]n" - Channel info(?)
    async def message_61_78(self, opcode, data):
        msg = FurcBuffer(data)
        version = msg.read220(1)
        channelUpdateType = msg.read220(1)
        unk1 = msg.read220(1)
        while msg.remaining >= 9:
            nameLength = msg.read220(1)
            name = msg.read(nameLength)
            bitpack = msg.read220(4)
            canJoin = bitpack >> 7 & 1
            maturity = (bitpack >> 8 & 0xf) * 100
            peopleCount = bitpack >> 25 & 0x1f
            canTell = bitpack & 0x4000
            canListen = bitpack & 0x20000
            canBroadcast = bitpack & 0x40000
            channelImage = msg.read220(4)
            await self.fire("ChannelInfo", channelUpdateType == "@", name, maturity, peopleCount, canTell, canListen, canBroadcast, channelImage)
    
    #"]o" - Dream owner name
    async def message_61_79(self, opcode, data):
        msg = FurcBuffer(data)
        owner = msg.read()
        await self.fire("DreamOwner", owner.decode())
    
    #"]q" - Load dream with custom patches
    async def message_61_81(self, opcode, data):
        msg = FurcBuffer(data)
        patchName = msg.readUntil()
        crc32 = msg.readUntil()
        unk1 = msg.readUntil()
        unk2 = msg.readUntil()
        modern = msg.readUntil()
        if modern == "modern":
            modern = True
        else:
            modern = False
        await self.fire("Dream", True, patchName, crc32, modern)
    
    #"]r" - Load dream with default patches
    async def message_61_82(self, opcode, data):
        msg = FurcBuffer(data)
        patchName = msg.readUntil()
        crc32 = msg.readUntil()
        unk1 = msg.readUntil()
        unk2 = msg.readUntil()
        modern = msg.readUntil()
        if modern == "modern":
            modern = True
        else:
            modern = False
        await self.fire("Dream", False, patchName, crc32, modern)
    
    #"]s" - Place text at location
    async def message_61_83(self, opcode, data):
        msg = FurcBuffer(data)
        x = msg.read95(2)
        y = msg.read95(2)
        textType = int(msg.readUntil()) #Text type: 0 = None, 1 = Normal, 2 = Lower
        name = msg.readUntil()
        title = msg.readUntil()
        maturity = int(msg.readUntil())
        gateType = int(msg.readUntil()) #Gate type: 0 = Normal, 1 = SS, 2 = Guild
        await self.fire("Text", (x, y), textType, name, title, maturity, gateType)
    
    #"]t" - Clear text at location
    async def message_61_84(self, opcode, data):
        msg = FurcBuffer(data)
        x = msg.read95(2)
        y = msg.read95(2)
        await self.fire("ClearText", (x, y))
    
    #"]u" - Waiting for dream
    #FIXME: Has a string parameter like 0_11, find out what this does
    async def message_61_85(self, opcode, data):
        await self.fire("UploadReady")
    
    #"]v" - Effect
    async def message_61_86(self, opcode, data):
        msg = FurcBuffer(data)
        #Known types:
        #a = breath
        #b = flame
        #c = glamour
        #d = splash
        effectType = msg.read(1)
        x = msg.read95(2)
        y = msg.read95(2)
        await self.fire("Effect",
            effectType.decode(), #Type
            (x, y) #Location
        )
    
    #"]w" - Version request
    async def message_61_87(self, opcode, data):
        await self.fire("VersionReq")
    
    #"]x" - Client should update
    async def message_61_88(self, opcode, data):
        await self.fire("Update")
    
    #"]y" - Changing IP/port
    async def message_61_89(self, opcode, data):
        await self.fire("Update", data)
    
    #"]z" - UID response
    async def message_61_90(self, opcode, data):
        msg = FurcBuffer(data)
        fuid = int(msg.readUntil())
        await self.fire("UID", fuid)
    
    #"]{" - Session ID
    async def message_61_91(self, opcode, data):
        msg = FurcBuffer(data)
        sessionID = int(msg.readUntil())
        await self.fire("SessionID", sessionID)
    
    #"]|" - Flip screen
    async def message_61_92(self, opcode, data):
        msg = FurcBuffer(data)
        flip = msg.read(1) == 49
        await self.fire("FlipScreen", flip)
    
    #"]}" - Set own colors
    async def message_61_93(self, opcode, data):
        msg = FurcBuffer(data)
        colors = Colors.fromStream(msg)
        await self.fire("SetColors", colors)
    
    #"^" - Item in hand
    async def message_62(self, opcode, data):
        msg = FurcBuffer(data)
        item = msg.read95(2)
        await self.fire("ButlerPaws", item)
    
    #"~" - Suspend drawing
    async def message_94(self, opcode, data):
        await self.fire("Suspend", True)
    
    #Hook definitions
    async def fire(self, name, *args, **kwargs):
        if name in self.listeners:
            for callback in self.listeners[name]:
                await callback(*args, **kwargs)
        
        for callback in self.listeners["*"]:
            await callback(name, *args, **kwargs)
        
        return True
    
    def on(self, name, *args, **kwargs):
        def _(func):
            if name not in self.listeners:
                self.listeners[name] = []
            self.listeners[name].append(func)
            return func
        return _
    
    def hook(self, name, func):
        if name not in self.listeners:
            self.listeners[name] = []
        self.listeners[name].append(func)
    
    def off(self, name, func):
        if name not in self.listeners:
            return
        self.listeners[name].remove(func)
    
    async def handlePacket(self, data):
        await self.fire("Raw", data)
        try:
            await getattr(
                self,
                "message_" + str(data[0]-32),
                self.message_unhandled
            )(data[0]-32, data[1:])
        except ValueError as e:
            print("DECODE FAILED ")

#Outgoing message definitions
class Commands:
    def login(self, character, machineid = None):
        if machineid == None:
            machineid = ""
        else:
            machineid = " "+machineid
        if character.type == character.TYPE_INI:
            return self.command("connect {} {}{}".format(
                character.name,
                character.password,
                machineid
            ))
        
        elif character.type == character.TYPE_ACCOUNT:
            return self.command("account {} {} {}{}".format(
                character.account.username,
                character.name,
                character.account.password,
                machineid
            ))
    
    def move(self, direction):
        return self.command("m {}".format(direction))
    
    def rotate(self, direction):
        if direction == 1:
            return self.command(">")
        
        elif direction == -1:
            return self.command("<")
    
    def vascodagama(self):
        return self.command("vascodagama")
    
    def say(self, message):
        return self.command("\"" + message)
    
    def gomap(self, mapid):
        return self.command("gomap " + encoding.b95encode(mapid))
    
    def fdl(self, url):
        return self.command("fdl "+url)
    
    def goToDream(self, furrename, dreamname = None):
        #TODO: URL Escape the names, could probably just shortname them
        url = "furc://" + furrename
        if dreamname:
            url += ":" + dreamname
        return self.fdl(url)

class Client(PacketHooks, Commands):
    def __init__(self, server = None):
        super().__init__()
        self.reader = None
        self.writer = None
        self.server = server or LIVE_SERVER
    
    #Basic networking stuff
    async def connect(self, server = None, loop = None, timeout = 5):
        if self.connected:
            await self.disconnect()
        
        if not server:
            server = self.server
        
        self.reader, self.writer = await asyncio.open_connection(
            server[0], server[1])
        
        motd = ""
        while self.connected:
            try:
                data = await asyncio.wait_for(self.reader.readline(), timeout)
            except asyncio.TimeoutError:
                data = None
                pass
            
            if not data: #None = Disconnected
                break
            
            if data[-1] != 10: #No EOL means incomplete stream + disconnected
                break
            
            if data == b"Dragonroar\n":
                return motd
            else:
                motd += data.decode()
        
        return None
    
    async def disconnect(self):
        writer.close()
        await writer.wait_closed()
        self.reader = None
        self.writer = None
    
    async def send(self, data):
        if self.connected:
            self.writer.write(data)
            await self.writer.drain()
            return True
        return False
    
    def command(self, data):
        if type(data) == str:
            data = data.encode()
        return self.send(data + b"\n")
    
    #async def handlePacket(self, data):
    #    pass
    
    @property
    def connected(self):
        if self.reader == None or self.writer == None:
            return False
        return True
    
    #Actual read loop, it is designed to be it's own task
    async def run(self):
        while self.connected:
            data = await self.reader.readline()
            
            if not data: #None = Disconnected
                break
            
            if data[-1] != 10: #No EOL means incomplete stream + disconnected
                break
            
            data = data[:-1] #Strip EOL
            
            if len(data) == 0: #If empty, ignore
                continue
            
            await self.handlePacket(data)
        
        #We are out of the loop! Presume Disconnected!
        self.reader = None
        self.writer = None



#!/usr/bin/env python3
from .colors import Colors
from .exceptions import LoginError
import datetime
import urllib.request
import urllib.parse
import urllib.error
import xml.etree.ElementTree as ET

LOGIN_SERVER = "https://charon.furcadia.com/accounts/clogin.php"

class Costume:
    def __init__(self, character, id, name, ordinal = None, colors = None,
                 lastPortrait = 0, lastScale = 100, lastGloam = 0,
                 lastSpecitag = 0, standard = 200, desc = ""):
        self.character = character
        self.id = id
        self.name = name
        self.ordinal = ordinal
        self.colors = colors
        self.lastPortrait = lastPortrait
        self.lastScale = lastScale
        self.lastGloam = lastGloam
        self.lastSpecitag = lastSpecitag
        self.standard = standard
        self.desc = desc

class Character:
    TYPE_INI = 0
    TYPE_ACCOUNT = 1
    
    def __init__(self):
        self.name = ""
        self.id = 0
        self.account = None
        self.colors = None
        self.password = None
        self.desc = None
        self.logins = 0
        self.lastLogin = 0
        self.autoResponse = None
        self.autoResponseMessage = None
        self.AFKTime = None
        self.AFKMessage = None
        self.AFKDescription = None
        self.AFKPortrait = None
        self.DefaultPortrait = None
        self.AFKDisconnectTime = None
        self.species = None
        self.flags = None
        self.lastItem = None
        self.lastPort = None
        self.lastScale = None
        self.lastGloam = None
        self.lastSpeci = None
        self.messages = None
        self.time = None
        self.owner = None
        self.costumes = None
        self.type = self.TYPE_INI
    
    @classmethod
    def fromINI(self, name, colors = None, password = None, desc = None,
                logins = 0, lastLogin = 0, autoResponse = None,
                autoResponseMessage = None, AFKTime = None, AFKMessage = None,
                AFKDescription = None, AFKPortrait = None,
                DefaultPortrait = None, AFKDisconnectTime = None):
        self = cls()
        self.type = self.TYPE_INI
        return self
    
    @classmethod
    def fromAccount(cls, account, id, name, species = None, colors = None,
                    flags = None, created = None, lastLogin = None,
                    lastItem = None, lastPort = None, lastScale = 100,
                    lastGloam = None, lastSpeci = None, messages = None,
                    logins = 0, time = None, owner = None, desc = None,
                    costumes = None):
        self = cls()
        self.type = self.TYPE_ACCOUNT
        self.account = account
        self.id = id
        self.name = name
        self.species = species
        self.colors = colors
        self.flags = flags
        self.created = created
        self.lastLogin = lastLogin
        self.lastItem = lastItem
        self.lastPort = lastPort
        self.lastScale = lastScale
        self.lastGloam = lastGloam
        self.lastSpeci = lastSpeci
        self.messages = messages
        self.logins = logins
        self.time = time
        self.owner = owner
        self.desc = desc
        self.costumes = costumes or []
        return self
    
    def addCostume(self, costume):
        if self.costumes:
            self.costumes.append(costume)

class Account:
    def __init__(self, username, password, id, email = None):
        self.username = username
        self.password = password
        self.id = id
        self.email = email
        self.characters = []
        pass
    
    def addCharacter(self, char):
        self.characters.append(char)
    
    def findCharacter(self, charname):
        for character in self.characters:
            if character.name.lower() == charname.lower(): #TODO: Compare by shortname
                return character
        return None
    
    @classmethod
    def login(cls, username, password, loginServer = None):
        data = urllib.parse.urlencode({
            "u": username,
            "p": password,
            "k": "ZAWJuuaTpRrG-Furcadia-KFzsVWwpPM9t",
            "v": "libfurc"
        })
        
        #TODO: Make this async
        req = urllib.request.Request(
            loginServer or LOGIN_SERVER,
            data = data.encode(),
            headers = {
                "Content-Type": "application/x-www-form-urlencoded"
            },
            method = "POST"
        )
        
        try:
            with urllib.request.urlopen(req) as res:
                root = ET.fromstring(res.read())
                account = None
                #Find the account entry (Future proofing)
                for child in root:
                    if child.tag == "account":
                        account = child
                        break
                
                if account == None:
                    raise exceptions.LoginError("No account element located!")
                
                self = cls(
                    username,
                    password,
                    int(account.attrib["id"]),
                    account.attrib.get("email", None)
                )
                
                for character in account:
                    char = Character.fromAccount(
                        self,
                        id = character.attrib["id"],
                        name = character.attrib["name"],
                        species = character.attrib["species"],
                        colors = Colors.fromCode(character.attrib["colors"]),
                        flags = int(character.attrib["flags"]),
                        created = datetime.datetime.strptime(character.attrib["created"], "%Y-%m-%d %H:%M:%S"),
                        lastLogin = datetime.datetime.strptime(character.attrib["lastlogin"], "%Y-%m-%d %H:%M:%S"),
                        lastItem = int(character.attrib["lastitem"]),
                        lastPort = int(character.attrib["lastport"]),
                        lastScale = int(character.attrib["lastscale"]),
                        lastGloam = int(character.attrib["lastgloam"]),
                        lastSpeci = int(character.attrib["lastspeci"]),
                        messages = int(character.attrib["messages"]),
                        logins = int(character.attrib["logins"]),
                        time = int(character.attrib["time"]),
                        owner = character.attrib["owner"]
                    )
                    
                    for costume in character:
                        if costume.tag == "costume":
                            char.addCostume(Costume(
                                character = char,
                                id = int(costume.attrib["id"]),
                                name = costume.attrib["name"],
                                ordinal = int(costume.attrib["ordinal"]),
                                colors = Colors.fromCode(costume.attrib["colors"]),
                                lastPortrait = int(costume.attrib["last_portrait"]),
                                lastScale = int(costume.attrib["last_scale"]),
                                lastGloam = int(costume.attrib["last_gloam"]),
                                lastSpecitag = int(costume.attrib["last_specitag"]),
                                standard = int(costume.attrib["standard"]),
                                desc = costume.text
                            ))
                        elif costume.tag == "desc":
                            char.desc = costume.text
                    
                    self.addCharacter(char)
            
        except urllib.error.HTTPError as e:
            raise LoginError(e.read()) from None
        
        return self


#!/usr/bin/env python3 
from . import base

class FurcBuffer:
    def __init__(self, buffer):
        self.buffer = buffer
        self.offset = 0

    @property
    def eof(self):
        return self.offset >= len(self.buffer)
    
    @property
    def remaining(self):
        return len(self.buffer) - self.offset
    
    def read(self, l = None):
        if l == None:
            v = self.buffer[self.offset:]
            self.offset = len(self.buffer)
        else:
            v = self.buffer[self.offset:self.offset+l]
            self.offset += l
        return v
    
    def readBuffer(self, l = 1):
        return FurcBuffer(self.read(l))
    
    def readUntil(self, seperator = b" "):
        data = b""
        while not self.eof:
            c = self.read(1)
            if c == seperator:
                return data
            data += c
        return data
    
    def read95(self, l = 1):
        try:
            return base.b95decode(self.read(l))
        except ValueError as e:
            raise ValueError("Failed to decode message {} at {}".format(self.buffer, self.offset - l))
    
    def read220(self, l = 1):
        try:
            return base.b220decode(self.read(l))
        except ValueError as e:
            raise ValueError("Failed to decode message {} at {}".format(self.buffer, self.offset - l))
    
    def read95Bytes(self, l = 1):
        try:
            return self.read(base.b95decode(self.read(l)))
        except ValueError as e:
            raise ValueError("Failed to decode message {} at {}".format(self.buffer, self.offset - l))
    
    def read95String(self, l = 1):
        return self.read95Bytes(l).decode()
    
    def read220Bytes(self, l = 1):
        try:
            return self.read(base.b220decode(self.read(l)))
        except ValueError as e:
            raise ValueError("Failed to decode message {} at {}".format(self.buffer, self.offset - l))
    
    def read220ByteArray(self, l = 1):
        result = []
        for i in range(l):
            v = self.read220(self.read(2))
            if v > 255: #No need to check for underflow, base220 is unsigned.
                raise ValueError("Base 220 byte array in message {} at {} is out of range!".format(self.buffer, self.offset - 2))
            result.append(v)
        
        return bytes(result)
    
    def read220String(self, l = 1):
        return self.read220Bytes(l).decode()


from distutils.core import setup
from libfurc import __version__ as version

long_description = None
with open("readme.md", "r") as f:
    long_description = f.read()

setup(
    name="libfurc",
    version=version,
    description="A library for handling Furcadia things.",
    long_description=long_description,
    long_description_content_type="text/markdown" if long_description else None,
    url="https://github.com/FelixWolf/libFurc",
    author="Félix",
    author_email="felix.wolfz@gmail.com",
    packages=["libfurc"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers", 
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: zlib/libpng License",
        "Topic :: Software Development :: Libraries"
    ]
)
